﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Joseph Helm and Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Common;
using HeuristicLab.Collections;
using HeuristicLab.Problems.BinPacking;

namespace HeuristicLab.Encodings.PackingEncoding {
  [Item("BinPacking", "Represents a single-bin packing for a bin-packing problem.")]
  [StorableClass]
  public abstract class BinPacking<D,B,I>  : Item
    where D : class, IPackingPosition
    where B : PackingShape<D>
    where I : PackingShape<D>, IPackingItem {
    #region Properties
    [Storable]
    public ObservableDictionary<int, D> ItemPositions { get; private set; } 

    [Storable]
    public ObservableDictionary<int, I> ItemMeasures { get; private set; } // TODO: rename to items

    [Storable]
    public B BinMeasures { get; private set; } 

    [Storable]
    public SortedSet<D> ExtremePoints { get; protected set; }

    [Storable]
    protected Dictionary<int, List<int>> OccupationLayers { get; set; }

    #endregion Properties

    protected BinPacking(B binMeasures) : base() {   
      ItemPositions = new ObservableDictionary<int, D>();
      ItemMeasures = new ObservableDictionary<int, I>();
      BinMeasures = (B)binMeasures.Clone();
      OccupationLayers = new Dictionary<int, List<int>>();
      InitializeOccupationLayers(); // TODO
    }


    [StorableConstructor]
    protected BinPacking(bool deserializing) : base(deserializing) { }
    protected BinPacking(BinPacking<D,B,I> original, Cloner cloner)
      : base(original, cloner) {
      this.ItemPositions = new ObservableDictionary<int, D>(original.ItemPositions);
      this.ItemMeasures = new ObservableDictionary<int, I>(original.ItemMeasures);
        this.BinMeasures = (B)original.BinMeasures.Clone(cloner);
        this.OccupationLayers = new Dictionary<int, List<int>>(original.OccupationLayers);
    }
   
    protected abstract void GenerateNewExtremePointsForNewItem(I measures, D position);

    public abstract D FindExtremePointForItem(I measures, bool rotated, bool stackingConstraints);
    public abstract D FindPositionBySliding(I measures, bool rotated);

    public abstract void SlidingBasedPacking(ref List<int> sequence, ItemList<I> itemMeasures);
    public abstract void SlidingBasedPacking(ref List<int> sequence, ItemList<I> itemMeasures, Dictionary<int, bool> rotationArray);
    public abstract void ExtremePointBasedPacking(ref List<int> sequence, ItemList<I> itemMeasures, bool stackingConstraints);
    public abstract void ExtremePointBasedPacking(ref List<int> sequence, ItemList<I> itemMeasures, bool stackingConstraints, Dictionary<int, bool> rotationArray);

    public void PackItem(int itemID, I measures, D position) {
      ItemMeasures[itemID] = measures;
      ItemPositions[itemID] = position;
      ExtremePoints.Remove(position);
      foreach (int id in ItemMeasures.Select(x => x.Key))
        GenerateNewExtremePointsForNewItem(ItemMeasures[id], ItemPositions[id]);
      
       
      //GenerateNewExtremePointsForNewItem(measures, position);
      //OccupiedPoints.OccupyPoints(measures, position, itemID);

      AddNewItemToOccupationLayers(itemID, measures, position);
    }

    public double PackingDensity { 
      get {
        double result = 0;
        foreach (var entry in ItemMeasures)
          result += entry.Value.Volume;
        result /= BinMeasures.Volume;
        return result;
      }
    }


    public int PointOccupation(D position) {
      //foreach (var ipEntry in ItemPositions) {
      //  if (ItemMeasures[ipEntry.Key].EnclosesPoint(ipEntry.Value, position))
      //    return ipEntry.Key;
      //}
      foreach (var id in GetLayerItemIDs(position)) {
        if (ItemMeasures[id].EnclosesPoint(ItemPositions[id], position))
          return id;
      }
      return -1;
    }

    public bool IsPointOccupied(D position) {
      //foreach (var ipEntry in ItemPositions) {
      //  if (ItemMeasures[ipEntry.Key].EnclosesPoint(ipEntry.Value, position))
      //    return true;
      //}      
      foreach (var id in GetLayerItemIDs(position)) {
        if (ItemMeasures[id].EnclosesPoint(ItemPositions[id], position))
          return true;
      }
      return false;
    }
    public bool IsPositionFeasible(I measures, D position) {
      //In this case feasability is defined as following: 1. the item fits into the bin-borders; 2. the point is supported by something; 3. the item does not collide with another already packed item
      if (!BinMeasures.Encloses(position, measures))
        return false;

      //foreach (var ipEntry in ItemPositions) {
      //  if (ItemMeasures[ipEntry.Key].Overlaps(ipEntry.Value, position, measures))
      //    return false;
      //} 
      foreach (var id in GetLayerItemIDs(measures, position)) {
        if (ItemMeasures[id].Overlaps(ItemPositions[id], position, measures))
          return false;
      }

      return true;
    }
    public abstract int ShortestPossibleSideFromPoint (D position);
    public abstract bool IsStaticStable (I measures, D position);


    protected abstract void InitializeOccupationLayers();
    protected abstract void AddNewItemToOccupationLayers(int itemID, I measures, D position);
    protected abstract List<int> GetLayerItemIDs(D position);
    protected abstract List<int> GetLayerItemIDs(I measures, D position);
  }   
}
